UNIT FLIP;
{$r-,q-,i-,x+,opt+,d-,l-}
{
This is the Flip part of the graphics subsystem.
------------------------------------------------
Current Version:
         0.14 - 06-03-99 by JacmeT (jacmet@iname.com)

Purpose: 
         To handle the pageflip (that is
         the copying of the screenbuffer
         to the screen) This involves
         conversion between different
         color depths and RGB / BGR.
         If possible tripple buffering
         is used.

Functionality:
         + Vesa 2.0+ support (15/16/24/32 bit modes)
         + Vga RRGGGBBB fallback
         + Tripple / double-buffering
         + Scaling / colordepth conversion on-the-fly (VGA)

         (minus indicates not done yet, 
         plus indicates it is implemented and tested
         and question mark indicates implemented
         (perhaps partly) but not tested yet)
           
Comments:
        The use of defines are heavilly approved.
        As of this moment the following defines are
        defined:
        ALL           : All code is defined
        _VESA         : Vesa support included
        _VGA          : VGA support included
        _15BIT        : Support for 15bit modes included.
        _16BIT        : Support for 16bit modes included.
        _24BIT        : Support for 24bit modes included.
        _32BIT        : Support for 32bit modes included.
        _RGB          : Support for RGB modes included.
        _BGR          : Support for BGR modes included.
        TRIPPLEBUFFER : Use Tripple buffering if available
        SCALING       : Use scaling (currently only for VGA)
        PMODEINTERFACE: Use Protected mode interface.
        WAITFORRETRACE: Wait for Vertical retrace before performing pageflip  

History:
         24-01-99: First revision (0.01)
         --------------------------------
         This is the very first version.
         Not much is working at the moment
         - But it'll come!   

         30-01-99: Second revision (0.02)
         --------------------------------
         Now 24 bit RGB/BGR and 32 bit
         RGBA/BGRA is working. At this
         time they only uses double
         buffering, but this can easily
         be changed when VBE function 07h
         is implemented in the VESA unit.
         Now the flip routines also 
         automatically adapts to different
         screen modes than the requested
         one, and handles VESA bios
         implementations with certain flaws
         (that is bytes per scanline 
         different from Xres*BitsPerPixel/8)        

         02-02-99: Third revision (0.03)
         --------------------------------
         Tripple buffering implemented!
         At this time the Pmode interface
         isn't implemented in the VESA
         unit, so only the int 10h interface
         works. WAITFORRETRACE added. When this
         is defined the tripplebuffering code
         will wait until the next vertical
         retrace before doing the pageflip.
         This should eliminate all flickering
         at the cost of speed.

         10-02-99: Forth revision (0.04)
         --------------------------------
         A framecounter variable added. This
         variable is increased every pageflip
         and can be used to calculate average
         framerate.

         14-02-99: Fifth revision (0.05)
         --------------------------------
         Minor cosmetic changes. Now all defines
         are done in FLIP.INC.
         Possible bug in 32 bit modes! - It seems
         that the 32bit modes are buggy (atleast
         on one gfx card (It could of cause also
         be a buggy VESA implementation on the
         card) - I'll have to look into that
         further.)

         17-02-99: Sixth revision (0.06)
         --------------------------------
         VGA flip routine added. VGA mode
         is changed from grayscale to
         RRRGGGBB color (3 bit red, 3 bit
         green and 2 bit blue)

         18-02-99: Seventh revision (0.07)
         ----------------------------------
         VGA flip routine updated. Now
         scaling is implemented,which means
         that modes different from 320x200
         will also work with VGA.
         X- and Y-resolution of virtual
         screen moved from VESA.PAS to 
         FLIP.PAS. This is done because 
         they are used both by VESA Flip
         procedures and the VGA Flip procedure
         with scaling.
         The Old (non scaling) version of the
         VGA Flip procedure can be found under
         the VGA_NOSCALING define. This will
         be removed as soon as the scaling
         version has been thoroughly tested.
         15 and 16bit RGB/BGR modes added.

         19-02-99: Eighth revision (0.08)
         ---------------------------------
         VGA_NOSCALING define removed. Instead
         the SCALING define is added.

         23-02-99: Ninth revision (0.09)
         --------------------------------
         Byte order changed! Now it's BGRA instead of
         RGBA. This is done because about 95% of all
         gfx cards have their truecolor modes aranged
         in BGR not RGB, and therefor it's inefficient
         to convert the byteorder for every pageflip.

         24-02-99: Tenth revision (0.10)
         --------------------------------    
         Added the _ASSEMBLER define. When
         this is defined, optimized (more
         or less) assembler versions of
         the pageflips will be used instead
         of the pascal versions. At this
         time only the Flip32bitBGR procedure
         has been converted.

         25-02-99: Yet another revision (0.11)
         --------------------------------------    
         This is mainly a bugfix release. I just
         discovered that I copied 1 more pixel
         than I should. that is, the pageflips
         were from 0..X*Y and not 0..(X*Y)-1.
         Also fixed a bug with the assembler
         version of Flip32bitRGB, which only
         copied 1/4 of the page (stupid me, 
         getting too used to the highlevel way
         of adressing one dword at a time.
         Now the sequence (eax=Flip.X,ecx=Flip.Y)
           mul ecx
           add eax,esi     (add [Source])
           dec eax
         is changed to
           mul ecx
           lea eax,[esi+eax*4-1] 
         I just browsed http://developer.intel.com
         and found some source for 24bit->16bit
         conversions using MMX, so hopefully I'll
         be able to implement those ideas into
         MMX optimized versions of the Pageflips.
         This requires some way of testing if MMX
         is supported at runtime, not just at
         compiletime (this can quite easily be 
         done using the CPUID function (after 
         checking whether it's supported) though).         

         28-02-99: Yet another revision (0.12)
         --------------------------------------
         Changed the order for Tripplebuffering.
         Earlier versions changed the visible 
         buffer before doing the pageflip. 
         Unfortunately this has the affect that 
         the image shown on the screen is delayed
         by one frame, and thus to show the very 
         first frame one has to do 2 pageflips -
         Therefor the order is now reversed. 
         First the virtual screen is copied to
         the buffer currently not shown, and 
         afterwards the buffers are flipped.

         03-03-99: Bugfix release (0.13)
         --------------------------------
         Fixed a bug in Assembler version
         of Flip32bitBGR (tripplebuffering
         code wasn't changed like the rest
         (see version 0.12))

         06-03-99: CA-II revision (0.14)
         --------------------------------
         I just discovered a cut-n-paste
         error in the Flip24BitRGB 
         procedure. The ending ENDIF of 
         the TRIPPLEBUFFER was part was
         unfortunately pasted as
         DEFINE TRIPPLEBUFFER, which 
         caused the GFX system to fail
         compiling without VESA support.
         Futher more I've started looking
         at the Hermes project (the 
         pageflip routines used in the
         PTC gfx system) - And I'm planning
         to implement some of the MMX
         optimized code from it into GFX
         system.
}

{$I HARDWARE.INC}
{$I ..\GENERAL.INC}

{$IFDEF ALL }
 {$DEFINE _VESA }
 {$DEFINE _VGA }
 {$DEFINE _RGB }
 {$DEFINE _BGR }
 {$DEFINE _32BIT }
 {$DEFINE _24BIT }
 {$DEFINE _16BIT }
 {$DEFINE _15BIT }
 {$DEFINE TRIPPLEBUFFER }
 {$DEFINE SCALING }
 {{$DEFINE PMODEINTERFACE }
 {$DEFINE WAITFORRETRACE }
{$ENDIF }

{$IFDEF _32BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _24BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _16BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _15BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFNDEF _BGR }
 {$DEFINE _RGB }
{$ENDIF }

{$IFNDEF _RGB }
 {$DEFINE _BGR }
{$ENDIF }

INTERFACE

{$IFDEF _VESA}
 {$IFDEF _VGA}
  USES VESA,VGA;
 {$ELSE}
  USES VESA;
 {$ENDIF}
{$ELSE}
 {$IFDEF _VGA}
  USES VGA;
 {$ENDIF}
{$ENDIF}

{$IFDEF _32BIT}
 {$IFDEF _RGB}
  Procedure Flip32bitBGR;
 {$ENDIF}
 {$IFDEF _BGR}
  {$IFDEF _ASSEMBLER}
   Procedure Flip32bitRGB;
  {$ELSE}
   Procedure Flip32bitRGB;
  {$ENDIF}
 {$ENDIF}
{$ENDIF}

{$IFDEF _24BIT}
 {$IFDEF _RGB}
  Procedure Flip24bitRGB;
 {$ENDIF}
 {$IFDEF _BGR}
  Procedure Flip24bitBGR;
 {$ENDIF}
{$ENDIF}

{$IFDEF _16BIT}
 {$IFDEF _RGB}
  Procedure Flip16bitRGB;
 {$ENDIF}
 {$IFDEF _BGR}
  Procedure Flip16bitBGR;
 {$ENDIF}
{$ENDIF}

{$IFDEF _15BIT}
 {$IFDEF _RGB}
  Procedure Flip15bitRGB;
 {$ENDIF}
 {$IFDEF _BGR}
  Procedure Flip15bitBGR;
 {$ENDIF}
{$ENDIF}

{$IFDEF _VGA} 
 Procedure FlipVGA;
{$ENDIF}

{$IFDEF _VESA}
var Buffer : array[0..1] of ^array[0..10000000] of dword;
    Address: array[0..1] of dword;
{$ENDIF}
var FrameCounter: dword;
    Source: ^array[0..10000000] of dword;
    FlipRoutine: ^Procedure;
    X,Y: dword;

IMPLEMENTATION

{$IFDEF _32BIT}
 {$IFDEF _BGR}
  {$IFDEF _ASSEMBLER}
   Procedure Flip32bitBGR; code;
    asm
     push esi
     push edi
     mov eax,[FLIP.FrameCounter]
     mov esi,[FLIP.Source]           {esi=[FLIP.Source]}
     mov ebx,[FLIP.X]
     inc eax                         {FLIP.FrameCounter+=1}
     mov ecx,ebx                     {ecx=FLIP.X}
     shl ebx,2                       {to get number of bytes (32bit)}
     movzx edx,[VESA.ModeInfo.BytesPerScan] {movzx because 16bit variable}
     mov [FLIP.FrameCounter],eax
     sub edx,ebx                     {edx=DeltaLine (number of bytes to skip at end of line)}

     {get destinaton buffer}
     {$IFDEF TRIPPLEBUFFER}
      and eax,1
      mov edi,dword ptr [FLIP.Buffer+eax*4]
     {$ELSE}
      mov edi,dword ptr [FLIP.Buffer]
     {$ENDIF}

     mov eax,[FLIP.Y]      
     mov ebx,ecx                     {ebx=FLIP.X}
     mul ecx                         {eax=FLIP.X*FLIP.Y}
     lea eax,[esi+eax*4-1]
     
    @nextline:
     rep movsd
     add edi,edx                     {source+=deltaline}
     mov ecx,ebx
     cmp esi,eax                     {are we done?}
     jb @nextline                    {if not then do nextline}
  
     {$IFDEF TRIPPLEBUFFER}
      mov eax,[FLIP.FrameCounter]
      and eax,1
      {$IFDEF PMODEINTERFACE}
       {$ERROR Pmode interface not implemented yet}
      {$ELSE}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF}        
       mov edx,dword ptr [FLIP.Address+eax*4]
       mov eax,4F07h {VBE function Set/Get Display Start}
       xor ecx,ecx
       int 10h
      {$ENDIF}
     {$ENDIF}

     {cleanup}
     pop edi
     pop esi
     ret
    end; {Flip32bitBGR}    
  {$ELSE}
   Procedure Flip32bitBGR;
    var s,d,EndOfLine,DeltaLine: dword;
    begin
     FrameCounter:=FrameCounter+1;     
     s:=0; {pointer in source}
     d:=0; {pointer in destination}
     EndOfLine:=FLIP.X;
     DeltaLine:=(dword(VESA.ModeInfo.BytesPerScan)-FLIP.X*4) shr 2;
     repeat
      repeat
       Buffer[FrameCounter and 1]^[d]  :=Source^[s];
       Buffer[FrameCounter and 1]^[d+1]:=Source^[s+1];
       Buffer[FrameCounter and 1]^[d+2]:=Source^[s+2];
       buffer[FrameCounter and 1]^[d+3]:=Source^[s+3];
       s:=s+4;
       d:=d+4;
      until (s>=EndOfLine);
      EndOfLine:=EndOfLine+FLIP.X;
      d:=d+DeltaLine;
     until s>=((FLIP.X*FLIP.Y)-1);
     {$IFDEF TRIPPLEBUFFER}
      {$IFDEF PMODEINTERFACE}
       {not implemented yet}
       VESA.DeInit;
       Writeln('FLIP: Pmode interface not supported yet');
       halt;
      {$ELSE}
       asm
        mov eax,4F07h {VBE function Set/Get Display Start}
        {$IFDEF WAITFORRETRACE}
         mov ebx,080h
        {$ELSE}
         xor ebx,ebx
        {$ENDIF} 
        mov ecx,[FrameCounter]
        and ecx,1
        mov edx,dword ptr [Address+ecx*4]
        xor ecx,ecx
        int 10h
       end; 
      {$ENDIF}
     {$ENDIF}
    end; {Flip32bitBGR}
  {$ENDIF}
 {$ENDIF}
 {$IFDEF _RGB}
  Procedure Flip32bitRGB;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;    
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(VESA.ModeInfo.BytesPerScan-FLIP.X*4) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=Source^[s] and $FF00FF00 + (Source^[s] and $000000FF) shl 16 + (Source^[s] and $00FF0000) shr 16;
      Buffer[FrameCounter and 1]^[d+1]:=Source^[s+1] and $FF00FF00 + (Source^[s+1] and $000000FF) shl 16 + (Source^[s+1] and $00FF0000) shr 16;
      Buffer[FrameCounter and 1]^[d+2]:=Source^[s+2] and $FF00FF00 + (Source^[s+2] and $000000FF) shl 16 + (Source^[s+2] and $00FF0000) shr 16;
      Buffer[FrameCounter and 1]^[d+3]:=Source^[s+3] and $FF00FF00 + (Source^[s+3] and $000000FF) shl 16 + (Source^[s+3] and $00FF0000) shr 16;
      s:=s+4;
      d:=d+4;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}
  end; {Flip32bitRGB}
 {$ENDIF}
{$ENDIF}

{ rem'ed out so far - will be implemented ASAP
Procedure Flip24bitBGR; code;
 asm
     push	ebp
      mov	ebp, esp
      and	ebp, -8
      sub	esp, 20
      mov	eax, dword ptr [FRAMECOUNTER]
      inc	eax
      mov	dword ptr [FRAMECOUNTER], eax
      mov	eax, 20231
      xor	ebx, ebx
      mov	ecx, [FRAMECOUNTER]
      and	ecx, 1
      mov	edx, dword ptr [ADDRESS + ecx * 4]
      xor	ecx, ecx
      int	16
      mov	dword ptr [ebp - 4], 0
      mov	dword ptr [ebp - 8], 0
      mov	eax, dword ptr [X]
      mov	dword ptr [ebp - 12], eax
      mov	edx, dword ptr [X]
      movzx	eax, word ptr [VESA.MODEINFO.BYTESPERSCAN]
      imul	edx, 3
      sub	eax, edx
      shr	eax, 2
      mov	dword ptr [ebp - 16], eax
@L2:
@L5:
      mov	edx, dword ptr [ebp - 4]
      mov	eax, dword ptr [SOURCE]
      mov	eax, dword ptr [eax + edx * 4]
      and	eax, 16777215
      mov	ecx, dword ptr [ebp - 4]
      mov	edx, dword ptr [SOURCE]
      inc	ecx
      mov	edx, dword ptr [edx + ecx * 4]
      shl	edx, 24
      add	edx, eax
      mov	eax, dword ptr [FRAMECOUNTER]
      mov	ecx, dword ptr [ebp - 8]
      and	eax, 1
      mov	eax, dword ptr [BUFFER + eax * 4]
      mov	dword ptr [eax + ecx * 4], edx
      mov	edx, dword ptr [ebp - 4]
      mov	eax, dword ptr [SOURCE]
      inc	edx
      mov	eax, dword ptr [eax + edx * 4]
      shr	eax, 8
      and	eax, 65535
      mov	ecx, dword ptr [ebp - 4]
      mov	edx, dword ptr [SOURCE]
      add	ecx, 2
      mov	edx, dword ptr [edx + ecx * 4]
      shl	edx, 16
      add	edx, eax
      mov	eax, dword ptr [FRAMECOUNTER]
      mov	ecx, dword ptr [ebp - 8]
      and	eax, 1
      inc	ecx
      mov	eax, dword ptr [BUFFER + eax * 4]
      mov	dword ptr [eax + ecx * 4], edx
      mov	edx, dword ptr [ebp - 4]
      mov	eax, dword ptr [SOURCE]
      add	edx, 2
      mov	eax, dword ptr [eax + edx * 4]
      shr	eax, 16
      and	eax, 255
      mov	ecx, dword ptr [ebp - 4]
      mov	edx, dword ptr [SOURCE]
      add	ecx, 3
      mov	edx, dword ptr [edx + ecx * 4]
      shl	edx, 8
      add	edx, eax
      mov	eax, dword ptr [FRAMECOUNTER]
      mov	ecx, dword ptr [ebp - 8]
      and	eax, 1
      add	ecx, 2
      mov	eax, dword ptr [BUFFER + eax * 4]
      mov	dword ptr [eax + ecx * 4], edx
      mov	eax, dword ptr [ebp - 4]
      add	eax, 4
      mov	dword ptr [ebp - 4], eax
      mov	eax, dword ptr [ebp - 8]
      add	eax, 3
      mov	dword ptr [ebp - 8], eax
      mov	eax, dword ptr [ebp - 4]
      cmp	eax, dword ptr [ebp - 12]
      jb	 @L5
      mov	eax, dword ptr [ebp - 12]
      add	eax, dword ptr [X]
      mov	dword ptr [ebp - 12], eax
      mov	eax, dword ptr [ebp - 8]
      add	eax, dword ptr [ebp - 16]
      mov	dword ptr [ebp - 8], eax
      mov	eax, dword ptr [X]
      imul	eax, dword ptr [Y]
      mov	edx, dword ptr [ebp - 4]
      cmp	edx, eax
      jb	 @L2
@L1:
      add	esp, 20
      pop	ebp
      ret	
end;
}
{$IFDEF _24BIT}
 {$IFDEF _BGR}
  Procedure Flip24bitBGR;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(dword(VESA.ModeInfo.BytesPerScan)-FLIP.X*3) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=Source^[s] and $00FFFFFF + Source^[s+1] shl 24;
      Buffer[FrameCounter and 1]^[d+1]:=(Source^[s+1] shr 8) and $0000FFFF + Source^[s+2] shl 16;
      Buffer[FrameCounter and 1]^[d+2]:=(Source^[s+2] shr 16) and $000000FF + Source^[s+3] shl 8;
      s:=s+4;
      d:=d+3;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}  
   end; {Flip24bitBGR}
 {$ENDIF}
 {$IFDEF _RGB}
  Procedure Flip24bitRGB;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(VESA.ModeInfo.BytesPerScan-FLIP.X*3) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=(Source^[s] shl 16) and $00FF0000 + (Source^[s] and $0000FF00) + (Source^[s] shr 16) and $000000FF + (Source^[s+1] shl 8) and $FF000000;
      Buffer[FrameCounter and 1]^[d+1]:=(Source^[s+1] shr 8) and $000000FF + (Source^[s+1] shl 8) and $0000FF00 + Source^[s+2] and $00FF0000 + (Source^[s+2] shl 16) and $FF000000;
      Buffer[FrameCounter and 1]^[d+2]:=Source^[s+2] and $000000FF + (Source^[s+3] shr 8) and $0000FF00 + (Source^[s+3] shl 8) and $00FF0000 + (Source^[s+3] shl 24);
      s:=s+4;
      d:=d+3;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}
   end; {Flip24bitRGB}
 {$ENDIF}
{$ENDIF}

{$IFDEF _16BIT}
 {$IFDEF _BGR}
  Procedure Flip16bitBGR;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(VESA.ModeInfo.BytesPerScan-FLIP.X*2) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=(Source^[s] and $000000F8) shr 3 + (Source^[s] and $0000FC00) shr 5 + (Source^[s] and $00F80000) shr 8
                                       +(Source^[s+1] and $000000F8) shl 13 + (Source^[s+1] and $0000FC00) shl 11 + (Source^[s+1] and $00F80000) shl 8;
      Buffer[FrameCounter and 1]^[d+1]:=(Source^[s+2] and $000000FF) shr 3 + (Source^[s+2] and $0000FC00) shr 5 + (Source^[s+2] and $00F80000) shr 8
                                       +(Source^[s+3] and $000000F8) shl 13 + (Source^[s+3] and $0000FC00) shl 11 + (Source^[s+3] and $00F80000) shl 8;
      s:=s+4;
      d:=d+2;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}
   end; {Flip16bitBGR}
 {$ENDIF}
 {$IFDEF _RGB}
  Procedure Flip16bitRGB;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(VESA.ModeInfo.BytesPerScan-FLIP.X*2) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=(Source^[s] and $000000F8) shl 8 + (Source^[s] and $0000FC00) shr 5 + (Source^[s] and $00F80000) shr 19 +
                                       +(Source^[s+1] and $000000F8) shl 24 + (Source^[s+1] and $0000FC00) shl 11 + (Source^[s+1] and $00F80000) shr 3;
      Buffer[FrameCounter and 1]^[d+1]:=(Source^[s+2] and $000000F8) shl 8 + (Source^[s+2] and $0000FC00) shr 5 + (Source^[s+2] and $00F80000) shr 19 +
                                       +(Source^[s+3] and $000000F8) shl 24 + (Source^[s+3] and $0000FC00) shl 11 + (Source^[s+3] and $00F80000) shr 3;
      s:=s+4;
      d:=d+2;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}
   end; {Flip16bitRGB}
 {$ENDIF}
{$ENDIF}

{$IFDEF _15BIT}
 {$IFDEF _BGR}
  Procedure Flip15bitBGR;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(VESA.ModeInfo.BytesPerScan-FLIP.X*2) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=(Source^[s] and $000000F8) shr 3 + (Source^[s] and $0000F800) shr 6 + (Source^[s] and $00F80000) shr 9
                                       +(Source^[s+1] and $000000F8) shl 13 + (Source^[s+1] and $0000F800) shl 10 + (Source^[s+1] and $00F80000) shl 7;
      Buffer[FrameCounter and 1]^[d+1]:=(Source^[s+2] and $000000FF) shr 3 + (Source^[s+2] and $0000F800) shr 6 + (Source^[s+2] and $00F80000) shr 9
                                       +(Source^[s+3] and $000000F8) shl 13 + (Source^[s+3] and $0000F800) shl 10 + (Source^[s+3] and $00F80000) shl 7;
      s:=s+4;
      d:=d+2;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}
   end; {Flip15bitBGR}
 {$ENDIF}
 {$IFDEF _RGB}
  Procedure Flip15bitRGB;
   var s,d,EndOfLine,DeltaLine: dword;
   begin
    FrameCounter:=FrameCounter+1;
    s:=0; {pointer in source}
    d:=0; {pointer in destination}
    EndOfLine:=FLIP.X;
    DeltaLine:=(VESA.ModeInfo.BytesPerScan-FLIP.X*2) shr 2;
    repeat
     repeat
      Buffer[FrameCounter and 1]^[d]  :=(Source^[s] and $000000F8) shl 7 + (Source^[s] and $0000F800) shr 6 + (Source^[s] and $00F80000) shr 19 +
                                       +(Source^[s+1] and $000000F8) shl 23 + (Source^[s+1] and $0000F800) shl 10 + (Source^[s+1] and $00F80000) shr 3;
      Buffer[FrameCounter and 1]^[d+1]:=(Source^[s+2] and $000000F8) shl 7 + (Source^[s+2] and $0000F800) shr 6 + (Source^[s+2] and $00F80000) shr 19 +
                                       +(Source^[s+3] and $000000F8) shl 23 + (Source^[s+3] and $0000F800) shl 10 + (Source^[s+3] and $00F80000) shr 3;
      s:=s+4;
      d:=d+2;
     until (s>=EndOfLine);
     EndOfLine:=EndOfLine+FLIP.X;
     d:=d+DeltaLine;
    until s>=((FLIP.X*FLIP.Y)-1);
    {$IFDEF TRIPPLEBUFFER}
     {$IFDEF PMODEINTERFACE}
      {not implemented yet}
      VESA.DeInit;
      Writeln('FLIP: Pmode interface not supported yet');
      halt;
     {$ELSE}
      asm
       mov eax,4F07h {VBE function Set/Get Display Start}
       {$IFDEF WAITFORRETRACE}
        mov ebx,080h
       {$ELSE}
        xor ebx,ebx
       {$ENDIF} 
       mov ecx,[FrameCounter]
       and ecx,1
       mov edx,dword ptr [Address+ecx*4]
       xor ecx,ecx
       int 10h
      end; 
     {$ENDIF}
    {$ENDIF}
   end; {Flip15bitRGB}
 {$ENDIF}
{$ENDIF}

{$IFDEF _VGA}
 {$IFDEF SCALING}
  Procedure FlipVGA;
  var s,d,EndOfLine,y,temp: dword;
      VGAscreen: array[0..15999] of dword absolute $A0000;
  begin
   FrameCounter:=FrameCounter+1;
   s:=0; d:=0; EndOfLine:=80; y:=0;
   repeat    
    repeat
     temp:=((Source^[s shr 8] and $E0) shr 5)+((Source^[s shr 8] and $E000) shr 10)+((Source^[s shr 8] and $C00000) shr 16);
     s:=s+VGA.DeltaX;
     temp:=temp+((Source^[s shr 8] and $E0) shl 3)+((Source^[s shr 8] and $E000) shr 2)+((Source^[s shr 8] and $C00000) shr 8);
     s:=s+VGA.DeltaX;
     temp:=temp+((Source^[s shr 8] and $E0) shl 11)+((Source^[s shr 8] and $E000) shl 6)+(Source^[s shr 8] and $C00000);
     s:=s+VGA.DeltaX;
     temp:=temp+((Source^[s shr 8] and $E0) shl 19)+((Source^[s shr 8] and $E000) shl 14)+((Source^[s shr 8] and $C00000) shl 8);
     s:=s+VGA.DeltaX;
     VGAscreen[d]:=temp;
     d:=d+1;
    until (d>=EndOfLine);
    EndOfLine:=EndOfLine+80;
    y:=y+DeltaY;
    S:=(y and $FFFFFF00)*FLIP.X;
   until d>15999;
  end; {FlipVGA}
 {$ELSE}
  Procedure FlipVGA;
  var s,d,temp: dword;
      VGAscreen: array[0..15999] of dword absolute $A0000;
  begin
   FrameCounter:=FrameCounter+1;
   s:=0; d:=0;
   repeat    
   temp:=((Source^[s] and $E0) shr 5)+((Source^[s] and $E000) shr 10)+((Source^[s] and $C00000) shr 16)+
         ((Source^[s+1] and $E0) shl 3)+((Source^[s+1] and $E000) shr 2)+((Source^[s+1] and $C00000) shr 8)+
         ((Source^[s+2] and $E0) shl 11)+((Source^[s+2] and $E000) shl 6)+(Source^[s+2] and $C00000)+
         ((Source^[s+3] and $E0) shl 19)+((Source^[s+3] and $E000) shl 14)+((Source^[s+3] and $C00000) shl 8);
   VGAscreen[d]:=temp;
   d:=d+1;
   s:=s+4;
   until d>15999;
  end; {FlipVGA}
 {$ENDIF}
{$ENDIF}
 
 BEGIN
 END.