UNIT GFX;
{$r-,q-,i-,x+,opt+,d-,l-}
{
This is the graphics subsystem main unit.
-----------------------------------------
Current Version:
         0.15 - 03-03-99 by JacmeT (jacmet@iname.com)

Purpose: 
         To handle the hardware interface
         to the gfx. card. In this way it
         is possible to write effects
         independent of the capabilities
         of the differnt graphics cards.
         With this tool the only thing the
         programmer has to do is to calculate
         the graphics in 32 bit truecolor (RGBA),
         and then call the FlipPage(Source: pointer)
         procedure.

Functionality:
         + Vesa 2.0+ support (15/16/24/32 bit modes)
         + Vga RRRGGGBB fallback
         + Tripple / double-buffering
         + Scaling / colordepth conversion on-the-fly

         (minus indicates not done yet, 
         plus indicates it is implemented and tested
         and question mark indicates implemented
         (perhaps partly) but not tested yet)
           
Comments:
        The use of defines are heavilly approved.
        As of this moment the following defines are
        defined:
        ALL           : All code is defined (Doesn't affect verbose level)
        _VESA         : Vesa support included
        _VGA          : VGA support included
        _15BIT        : Support for 15bit modes included.
        _16BIT        : Support for 16bit modes included.
        _24BIT        : Support for 24bit modes included.
        _32BIT        : Support for 32bit modes included.
        FRAMERATE     : Calculate average framerate
        VERBOSE_LEVEL0: Verbose level 0 - no messages are included
        VERBOSE_LEVEL1: Verbose level 1 - only the most important messages are included
        VERBOSE_LEVEL2: Verbose level 2 - most messages are included
        VERBOSE_LEVEL3: Verbose level 3 - all messages are included
        DEBUG         : Debugging code included
        ERRORHANDLING : Errors are handled.
  
History:
         18-01-99: First revision (0.01)
         --------------------------------
         This is the very first version.
         Not much is working at the moment
         - But it'll come!   

         25-01-99: Second revision (0.02)
         --------------------------------
         Added the Flip unit, which caused
         me to change the Flip procedure
         to _Flip. This is only temporary!
         Now the Init,Setmode,_Flip and
         DeInit seems to work - this means
         we can start testing! 
         Added the PixelType type for 
         programs to use.

         30-01-99: Third revision (0.03)
         --------------------------------
         Added support for 32 bit modes.
         Added a check for RGB/BGR - now
         GFX_SYSTEM should work on both
         kinds of truecolor modes.
         In the next revision I'll probably
         add support in the setmode 
         procedure for tripple buffering and
         the testing of bigger screenmodes
         (e.g. if the user request 512x384
         the system checks the following 
         resolutions: 512x384x32,512x384x24
         640x400x32,640x400x24,640x480x32...)
         until a supported mode is found.
         This can be done in a table of
         Possible X- and Y-resolutions.

         01-02-99: Forth revision (0.04)
         --------------------------------
         Changed the Setmode procedure.
         This is done because of the changes
         done in the VESA unit. Now the system
         automatically tries out larger 
         screenmodes if the requested one isn't
         supported.
         Changed the order of screenmodes to try
         out. Before the order were: 24/32/16/15.
         Now it's 32/24/16/15. This is done to
         Simplify the Setmode procedure / VESA 
         unit, and it might change in the future.

         02-02-99: Fifth revision (0.05)
         --------------------------------
         Tripple buffering finally
         implememted in Setmode procedure. 
         VERBOSE_LEVEL0 added. With 
         VERBOSE_LEVEL0 the GFX unit will
         never write any info on screen.

         14-02-99: Sixth revision (0.06)
         --------------------------------
         Minor cosmetic changes. Now all defines
         are done in GFX.INC - The system has now
         been under beta testing, and it seems 
         quite stable and bugfree. (Perhaps I've
         found a bug in the 32bit flip though -
         but I haven't had time to look into that
         further.)
         The _Flip procedure is renamed to FlipPage
         as promished in version 0.02.
         A public VERSION string (constant) added.

         17-02-99: Seventh revision (0.07)
         --------------------------------
         Cleaned up the Setmode code. Now it's
         cleaner, smaller and supports VGA fallback
         properly.
         Added full support for VGA fallback. Now
         GFX_SYS uses VGA mode 13h RRRGGGBB if VESA
         2.0+ isn't supported (with linear framebuffer),
         or if the requested mode isn't supported.

         18-02-99: Eighth revision (0.08)
         --------------------------------
         Minor restructuring of .INC files
         to avoid different defines in
         GFX.PAS and FLIP.PAS - now all
         hardware related is put in FLIP.INC
         which GFX.PAS also includes.
         Bugfix for Setmode: The restructuring
         of the Setmode procedure in last
         revision causes a 'GFX: No hardware
         support detected' error (if
         ERRORHANDLING is defined) if f.eks.
         _32BIT isn't defined, and the program is
         running on a computer which has hardware
         support for 32 bit modes. This is caused
         by the fact that the Setmode procedure
         detects the 32bit mode and tries to set
         it, but the code to handle a 32bit mode
         isn't defined, and therefor the system
         fails. This bugfix doesn't fix all bugs,
         because it just advanced to the next
         supported mode in the modelist - and if
         that isn't supported either, then we have
         the problem again. As soon as possible I
         will change the bugfix and perhaps 
         implement it in the GetSupportedModes call
         in VESA.PAS instead. This way the modes
         that aren't defined will not be inserted
         in the ModeList in the first place and we
         won't have the problem at all.
         But until then just include support for
         every mode and you'll be fine ($DEFINE ALL
         in FLIP.INC)         
         Added the Copyright Constant.

         18-02-99: Ninth revision (0.09)
         --------------------------------
         Just a quick hack for a performance
         counter. When the FRAMERATE define
         is defined the current bios clock
         tick (which is updated every 55 ms)
         is fetched as the last command in the
         SetMode procedure. Later when the
         GFX.DeInit procedure is called the
         average framerate is calculated using
         the framecounter variable from the
         FLIP unit.
         This is done by the following formula:
         Avg.Framerate=No.Frames/((Sstop-Sstart)*0.055)           

         19-02-99: Tenth revision (0.10)
         --------------------------------
         Version 0.10 reached! - The
         System seems quite stable now,
         even though I just discovered
         that it fails on a computer
         (with a ATI Mach64 gfx. card)
         that the earlier versions did
         work on - Perhaps it is the new
         revised Setmode procedure or
         just the fact that I moved the
         code base from TMT version 2.00
         and pmodew to version 2.11 and
         the new version 0.32 TMT stub
         (with the tmtpack .exe packer)
         - But I'll have to do a bit more
         testing on that. Futher more
         There seems to be something weird
         happening on a S3-TRI64V2 that I
         tested it on. I'll probably 
         implement the DEBUG define soon,
         which writes all the data related
         to the detected hardware and the 
         screenmode chosen to a file for
         debugging purposes. The reason for
         doing it this way (and not just 
         using the VERBOSE_LEVELX is that
         it helps debugging when we're in
         a gfx. mode (and hence, cannot use
         the standard write/writeln calls).
         A new file is added to the project
         'GFX_SYS.INF' - This is an info
         file about the structure of the 
         system and how to use it - This will
         have to do for a manual for now (heck,
         how many people is using it at this
         moment?)
         A copyright string is added, which may
         or may not later be merged with the
         VERSION constant.           

         21-02-99: Eleventh revision (0.11)
         -----------------------------------
         COPYRIGHT string merged with VERSION
         DATE.INC include added using the
         gendate utility from the p32 compiler
         (http://surf.to/p32). This way it is
         easy to see when the gfx system has
         been compiled.

         23-02-99: Yet-another revision (0.12)
         --------------------------------------
         Yet another minor restructoring of
         include files. It seems it wasn't that
         well planned to start with ;) - but
         now it seems ok. All hardware related
         defines are found in HARDWARE.INC 
         (basically what was found in FLIP.INC)
         This is included in VESA.PAS,FLIP.PAS
         and GFX.PAS (and perhaps later VGA.PAS)        
         This enabled me to implement the bugfix
         discussed in version 0.08.
         Changed the interface. I know this isn't
         exactly a good thing, since this may break
         programs using GFX_SYS, but it's done
         to make the interface even simpler. Now
         it isn't the calling program's responsibility
         to call Init and DeInit any longer. They are
         instead automaticly invoked by the 
         exitprocedure function and the begin end of a
         unit. This means that all the program has to
         do is to call the Setmode procedure with the
         desired resolution.
         GFX_SYS.INF updated to reflect changes.
         Added debug support. This should speed up
         remote debugging / beta testing. This way
         all data of the system is written to a file
         which can then be examined later. This way
         users who discovers a bug can send me a copy
         of the file, and it should help debugging
         quite a lot.
         PixelType Type changed! Now it's BGRA instead
         of RGBA. This is done because about 95% of all
         gfx cards have their truecolor modes aranged
         in BGR not RGB, and therefor it's inefficient
         to convert the byteorder for every pageflip.
         This will cause problems for programs filling
         the entire dword at a time (using the .COLOR
         instead of .R, .G and .B)

         25-02-99: Just-Do-It revision (0.13)
         -------------------------------------
         Bugfix release. I turned on range checking
         and the systems barfed at me. Various minor
         bugs related to stability under pure dos
         fixed (Mainly dealing with DPMI and out of
         bounds exceptions) - Now it should be stable
         again.

         28-02-99: Lets-MAKE-things-better (0.14)
         -----------------------------------------
         MakeFile created. Now the compilation and
         distribution of new versions are greatly
         simplified. For this to work I had to 
         create a new utility 'makeinc' which can 
         generate  include files with the chosen
         defines. Furtermore the include files has
         been restructured a bit to accommodate the
         changes.
         Changed the order for Tripplebuffering 
         (see FLIP.PAS)

         03-03-99: Bugfix release (0.15)
         --------------------------------
         Fixed a bug releated to the make
         file. The new utility 'MakeInc'
         would still recreate the include
         file even though the parameters
         hadn't changed. Also a bugfix
         is done in the Flip unit. 
         Started working on a C version
         using DJGPP.
}

{$I HARDWARE.INC}
{$I GFX.INC}
{$I ..\GENERAL.INC}

{$IFDEF ALL }
 {$DEFINE _VESA }
 {$DEFINE _VGA }
 {$DEFINE _RGB }
 {$DEFINE _BGR }
 {$DEFINE _32BIT }
 {$DEFINE _24BIT }
 {$DEFINE _16BIT }
 {$DEFINE _15BIT }
 {$DEFINE TRIPPLEBUFFER }
 {$DEFINE SCALING }
 {{$DEFINE PMODEINTERFACE }
 {$DEFINE WAITFORRETRACE }
{$ENDIF }

{$IFDEF _32BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _24BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _16BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFDEF _15BIT }
 {$DEFINE _VESA }
{$ENDIF }

{$IFNDEF _BGR }
 {$DEFINE _RGB }
{$ENDIF }

{$IFNDEF _RGB }
 {$DEFINE _BGR }
{$ENDIF }

INTERFACE

Const VERSION: String = '0.15 by Peter Korsgaard''99'; 
{$I DATE.INC} {to get the date-of-compilation}

Procedure Setmode(X,Y: dword);
{ The Setmode procedure.}
Procedure FlipPage(Source: Pointer);
{ The FlipPage procedure.}         
Procedure DeInit;
procedure Init;

Type PixelType = record
                  case integer of
                  0: (Color: dword);
                  1: (B,G,R,A: byte);
                 end;         

IMPLEMENTATION

{$IFDEF DEBUG}
 {$IFDEF _VESA}
  {$IFDEF _VGA}
   uses VESA,VGA,FLIP,DEBUGGER;
  {$ELSE}
   uses VESA,FLIP,DEBUGGER;
  {$ENDIF}  
 {$ELSE}
  {$IFDEF _VGA}
   uses VGA,FLIP,DEBUGGER;
  {$ELSE}
   {$ERROR either _VESA or _VGA has to be defined!}
  {$ENDIF}
 {$ENDIF}
{$ELSE}
 {$IFDEF _VESA}
  {$IFDEF _VGA}
   uses VESA,VGA,FLIP,DEBUG;
  {$ELSE}
   uses VESA,FLIP;
  {$ENDIF}
 {$ELSE}
  {$IFDEF _VGA}
   uses VGA,FLIP;
  {$ELSE}
   {$ERROR either _VESA or _VGA has to be defined!}
  {$ENDIF}  
 {$ENDIF}
{$ENDIF}

 
{$IFDEF FRAMERATE}
 const BIOSTICKADDRESS = $46C; {a bios tick is 1/18.2 seconds = 0.55 secs}
{$ENDIF}

var UseVesa: dword; {1 = use VESA, 2 = do not use VESA (bw mode 13h) - this is set in Init}
    OldExitProcedure: pointer; {to DeInit automatically}
{$IFDEF FRAMERATE}
var StartTime: dword;
{$ENDIF}

Procedure Init;
{ The initialization procedure. 
  -----------------------------
Purpose:
         This procedure detects the
         available hardware and sets
         up the gfx subsystem.

Comments:
         This procedure must be called
         before any of the others -
         See also the DeInit procedure
}         
 begin
  {Set FrameCounter = 0}
  Flip.FrameCounter:=0;

  {$IFDEF VERBOSE_LEVEL2}
   Writeln('GFX: Initializing system (version ',GFX.VERSION,')');
  {$ENDIF}

  UseVesa:=0; {to let us be able to know if we should use VESA}

  {$IFDEF _VESA}
   if (VESA.Init=True) then
    begin
     UseVesa:=1;     
     {$IFDEF DEBUG}
      Writeln(DEBUGGER.FileHandle,'GFX: Using VESA');
     {$ENDIF}
     VESA.GetSupportedModes;
     {$IFDEF VERBOSE_LEVEL3}
      Writeln('GFX: Using VESA');
     {$ENDIF}
    end;
  {$ENDIF}

  {$IFDEF _VGA}
   if (UseVesa=0) then {check if we didn't detect Vesa}
    begin
     if (VGA.Init=True) then
      begin
       UseVesa:=2;
       {$IFDEF DEBUG}
        Writeln(DEBUGGER.FileHandle,'GFX: Using VGA');
       {$ENDIF}
       {$IFDEF VERBOSE_LEVEL3}
        Writeln('GFX: Using VGA');
       {$ENDIF}
      end;
    end;
  {$ENDIF}
  
  {$IFDEF ERRORHANDLING}
   if (UseVesa=0) then
    begin
     {$IFDEF DEBUG}
      Writeln(DEBUGGER.FileHandle,'GFX: Failed to initialize hardware');
     {$ENDIF}
     {$IFDEF VERBOSE_LEVEL1}
      Writeln('GFX: Failed to initialize hardware');
     {$ENDIF}
     halt;
    end; 
  {$ENDIF}
 end;

Procedure Setmode(X,Y: dword);
{ The Setmode procedure. 
  -----------------------------
Purpose:
         This procedure tries to set
         up the grahics card in mode
         XxYxZbit. (where Z = the modes
         that are DEFINED) If this is not
         succesfull other modes will
         be tried (according to which
         are DEFINED)

Comments:
         This procedure affects the 
         flip pointer. That is, it
         installs the correct flip
         routine for the current
         screenmode.
         Remember to call Init before
         this call
}         
 var Count: longint;
     FoundMode: boolean;
 begin
  {$IFDEF DEBUG}
   Writeln(DEBUGGER.FileHandle,'GFX: Setmode ',X,'x',Y,' called');
  {$ENDIF}
  {$IFDEF VERBOSE_LEVEL2}
   Writeln('GFX: Setmode ',X,'x',Y,' called');
  {$ENDIF}
  {make FLIP.FlipRoutine point to nil 
   - this way we can check if a mode
   is found}
  FLIP.FlipRoutine:=nil;
  { Set size of virtual screen }
  FLIP.X:=X; FLIP.Y:=Y; 

  {$IFDEF _VESA}
   if (UseVesa=1) then
    begin
     Count:=-1;
     FoundMode:=False;
     {search through our modelist until
      we get to a mode which is equal
      or bigger than the requested one.
      This will always succeed, because
      the rest of the modelist is filled
      with X=0xFFFF, Y=0xFFFF.}
     repeat 
      Count:=Count+1;
     until ((VESA.ModeList[Count].X>=X) and (VESA.ModeList[Count].Y>=Y));

     repeat
      {$IFDEF DEBUG}
       Writeln(DEBUGGER.FileHandle,'GFX: Setmode ',VESA.ModeList[Count].X,'x',VESA.ModeList[Count].Y,'x',VESA.ModeList[Count].BitsPerPixel,' called');
      {$ENDIF}

      {$IFDEF VERBOSE_LEVEL3}
       {write info}
       {$IFDEF _32BIT}
        If VESA.ModeList[Count].BitsPerPixel=32 then 
         begin
          Writeln('GFX: Setmode ',VESA.ModeList[Count].X,'x',VESA.ModeList[Count].Y,'x32 called');
         end;
       {$ENDIF}
       {$IFDEF _24BIT}
        If VESA.ModeList[Count].BitsPerPixel=24 then 
         begin
          Writeln('GFX: Setmode ',VESA.ModeList[Count].X,'x',VESA.ModeList[Count].Y,'x24 called');
         end;
       {$ENDIF}
       {$IFDEF _16BIT}
        If VESA.ModeList[Count].BitsPerPixel=16 then 
         begin
          Writeln('GFX: Setmode ',VESA.ModeList[Count].X,'x',VESA.ModeList[Count].Y,'x16 called');
         end;
       {$ENDIF}
       {$IFDEF _15BIT}
        If VESA.ModeList[Count].BitsPerPixel=15 then 
         begin
          Writeln('GFX: Setmode ',VESA.ModeList[Count].X,'x',VESA.ModeList[Count].Y,'x15 called');
         end;
       {$ENDIF}
      {$ENDIF}

      {could we set mode? - just in case we have a buggy VBE}
      if VESA.SetMode(VESA.ModeList[Count].ModeNumber) then
       begin
        FoundMode:=True;
       end;

      if FoundMode=False then Count:=Count+1;
     until ((FoundMode=True) or (Count>255));

     if (FoundMode=True) then
      begin
       {$IFDEF DEBUG}
        Writeln(DEBUGGER.FileHandle,'GFX: Mode set successfully');
       {$ENDIF}
       {place screen in middle of actual screen}
       {the reason why it's (VESA.ModeInfo.Bitsperpixel+1)
        is because 15bit modes are 2 bytes big !(15 shr 3)=2}
       dword(FLIP.Buffer):=dword(VESA.ModeInfo.PhysBasePtr)+
         ((dword(VESA.ModeInfo.Yres)-FLIP.Y) shr 1)*dword(VESA.ModeInfo.BytesPerScan)+
         ((dword(VESA.ModeInfo.Xres)-FLIP.X) shr 1)*((dword(VESA.ModeInfo.Bitsperpixel)+1) shr 3);

       {$IFDEF DEBUG}
        Writeln(DEBUGGER.FileHandle,'GFX: Buffer[0] = ',dword(Buffer[0]));
       {$ENDIF}

       {$IFDEF TRIPPLEBUFFER}
        {do we have enough memory for tripple buffering?}
        if (VESA.ControllerInfo.Memory shl 15)>=dword(VESA.ModeInfo.BytesPerScan)*dword(VESA.ModeInfo.Yres) then
         begin
          {use tripple buffering}
          dword(FLIP.Buffer[1]):=dword(FLIP.Buffer[0])+dword(VESA.ModeInfo.BytesPerScan)*dword(VESA.ModeInfo.Yres);
          FLIP.Address[1]:=VESA.ModeInfo.Yres;
          {$IFDEF DEBUG}
           Writeln(DEBUGGER.FileHandle,'GFX: Using Tripplebuffering');
           Writeln(DEBUGGER.FileHandle,'GFX: Buffer[1] = ',dword(Buffer[1]));
           Writeln(DEBUGGER.FileHandle,'GFX: Address[1] = ',dword(Address[1]));
          {$ENDIF}
         end
        else  
         begin
          {don't use tripple buffering}
          dword(FLIP.Buffer[1]):=dword(FLIP.Buffer[0]);
          FLIP.Address[1]:=1; 
          {$IFDEF DEBUG}
           Writeln(DEBUGGER.FileHandle,'GFX: Not enough memory for Tripplebuffering');
           Writeln(DEBUGGER.FileHandle,'GFX: Buffer[1] = ',dword(Buffer[1]));
           Writeln(DEBUGGER.FileHandle,'GFX: Address[0] = ',dword(Address[0]));
          {$ENDIF}
         end;
       {$ELSE}  
        {don't use tripple buffering}
        dword(FLIP.Buffer[1]):=dword(FLIP.Buffer[0]);
        FLIP.Address[1]:=1; 
        {$IFDEF DEBUG}
         Writeln(DEBUGGER.FileHandle,'GFX: Do not use Tripplebuffering');
         Writeln(DEBUGGER.FileHandle,'GFX: Buffer[1] = ',dword(Buffer[1]));
         Writeln(DEBUGGER.FileHandle,'GFX: Address[1] = ',dword(Address[1]));
        {$ENDIF}
       {$ENDIF}
       FLIP.Address[0]:=0;
       {$IFDEF DEBUG}
        Writeln(DEBUGGER.FileHandle,'GFX: Address[0] = ',dword(Address[0]));
       {$ENDIF}

       {is it RGB or BGR?}
       if VESA.ModeInfo.RedFieldPos=0 then 
        begin
         {$IFDEF _RGB}
          {$IFDEF DEBUG}
           Writeln(DEBUGGER.FileHandle,'GFX: Mode is RGB');
          {$ENDIF}     
          {$IFDEF _32BIT}
           If VESA.ModeList[Count].BitsPerPixel=32 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip32bitRGB');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip32bitRGB;         
            end; 
          {$ENDIF} 
          {$IFDEF _24BIT}
           If VESA.ModeList[Count].BitsPerPixel=24 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip24bitRGB');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip24bitRGB;
            end;
          {$ENDIF} 
          {$IFDEF _16BIT}
           If VESA.ModeList[Count].BitsPerPixel=16 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip16bitRGB');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip16bitRGB;
            end;
          {$ENDIF} 
          {$IFDEF _15BIT}
           If VESA.ModeList[Count].BitsPerPixel=15 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip15bitRGB');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip15bitRGB;
            end;
          {$ENDIF} 
         {$ELSE}
          {$IFDEF ERRORHANDLING}
           asm mov ax,3h; int 10h; end;
           Writeln('GFX: RGB modes not supported');
           readln;
          {$ENDIF}
         {$ENDIF} 
       end
       else
        begin
         {$IFDEF _BGR}
          {$IFDEF DEBUG}
           Writeln(DEBUGGER.FileHandle,'GFX: Mode is BGR');
          {$ENDIF}
          {$IFDEF _32BIT}
           If VESA.ModeList[Count].BitsPerPixel=32 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip32bitBGR');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip32bitBGR;         
            end;
          {$ENDIF} 
          {$IFDEF _24BIT}
           If VESA.ModeList[Count].BitsPerPixel=24 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip24bitBGR');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip24bitBGR;
            end;
          {$ENDIF} 
          {$IFDEF _16BIT}          
           If VESA.ModeList[Count].BitsPerPixel=16 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip16bitBGR');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip16bitBGR;
            end;
          {$ENDIF} 
          {$IFDEF _15BIT}
           If VESA.ModeList[Count].BitsPerPixel=15 then 
            begin
             {$IFDEF DEBUG}
              writeln(DEBUGGER.FileHandle,'GFX: Using Flip15bitBGR');
             {$ENDIF}
             FLIP.FlipRoutine:=@FLIP.Flip15bitBGR;
            end;
          {$ENDIF} 
         {$ELSE}
          {$IFDEF ERRORHANDLING}
           asm mov ax,3h; int 10h; end;
           Writeln('GFX: BGR modes not supported');
           readln;
          {$ENDIF}
         {$ENDIF} 
        end;
      end
     else
      begin
       {$IFDEF DEBUG}
         Writeln(DEBUGGER.FileHandle,'GFX: Mode set failed');
       {$ENDIF}

       {$IFDEF _VGA}
        UseVesa:=2;
       {$ENDIF} 
      end;
    end;
  {$ENDIF}

  {$IFDEF _VGA}
   if (UseVesa=2) then
    begin
     {$IFDEF DEBUG}
       Writeln(DEBUGGER.FileHandle,'GFX: Setting VGA mode');
     {$ENDIF}
     VGA.Setmode(X,Y);
     FLIP.FlipRoutine:=@FLIP.FlipVGA;         
    end; 
  {$ENDIF} 
  {$IFDEF ERRORHANDLING}
   if FLIP.FlipRoutine=nil then
    begin
     {$IFDEF DEBUG}
      Writeln(DEBUGGER.FileHandle,'GFX: No hardware support detected');
     {$ENDIF}

     {$IFDEF VERBOSE_LEVEL1}
      Writeln('GFX: No hardware support detected');
     {$ENDIF}
     halt;
    end;
  {$ENDIF}

  {$IFDEF FRAMERATE}
   StartTime:=memd[BIOSTICKADDRESS]; {address of bios timer ticks since midnight}
   {$IFDEF DEBUG}
    Writeln(DEBUGGER.FileHandle,'GFX: StartTime = ',StartTime);
   {$ENDIF}
  {$ENDIF}
 end;

Procedure FlipPage(Source: Pointer);
{ The FlipPage procedure. 
  -----------------------------
}
 begin
  dword(FLIP.Source):=dword(Source);
  FLIP.FlipRoutine^;
 end;

Procedure DeInit;
{ The Deinitialization procedure. 
  -----------------------------
Purpose:
         This procedure cleans up the
         various features the gfx system
         may have used and shuts down the 
         system returning to standard 
         mode 3 textmode

Comments:
         This procedure must be called
         as the last procedure in the
         gfx subsystem
}         
 begin
  {$IFDEF _VESA}
   if UseVesa=1 then VESA.DeInit;
  {$ENDIF}

  {$IFDEF _VGA}
   if UseVesa=2 then VGA.DeInit;
  {$ENDIF}  

  {$IFDEF VERBOSE_LEVEL2}
   Writeln('GFX: DeInitialization successful');
  {$ENDIF}
  {$IFDEF FRAMERATE}
   {the bios timer tick is updated
    approx. every 0.055 (1/18.2) second}
   Writeln('GFX: Average framerate: ',FLIP.FRAMECOUNTER/((memd[BIOSTICKADDRESS]-StartTime)*0.055):5:5,' Frames/S');
   {$IFDEF DEBUG}
    Writeln(DEBUGGER.FileHandle,'GFX: EndTime = ',memd[BIOSTICKADDRESS]);
   {$ENDIF}
  {$ENDIF}
  Writeln('GFX: Version ',GFX.VERSION,' [',COMPILEDATE,']');
  {restore exitprocedure}
  exitproc:=OldExitprocedure;
 end;

 begin
  GFX.Init;
  OldExitprocedure:=exitproc;
  exitproc:=@GFX.DeInit;
 end.